"""Reconnecting websocket client"""
import asyncio
import logging
from typing import Coroutine

from websockets.asyncio.client import ClientConnection, connect, process_exception
from websockets.exceptions import ConnectionClosed
from websockets.protocol import State

from .task import TaskList, task_manager

_LOGGER = logging.getLogger(__name__)


class ReconnectingWebsocketClient:
    """Reconnecting websocket client"""

    def __init__(
        self,
        host: str,
        port: int,
        on_message: Coroutine[None, str, None],
        on_state_change: Coroutine[None, State, None] = None,
        on_connect: Coroutine[None, None, None] = None,
        connect_timeout=15,
    ):
        self._host = host
        self._port = port
        self._connect_timeout = connect_timeout
        self._outgoing_queue: asyncio.Queue[str] = asyncio.Queue()
        self._incoming_queue: asyncio.Queue[str] = asyncio.Queue()
        self._on_message = on_message
        self._on_state_change = on_state_change
        self._on_connect = on_connect
        self._tasks = TaskList()
        self._ws = None

    @property
    def state(self):
        """State of connection"""
        if self._ws:
            return self._ws.protocol.state

    async def _do_on_state_change(self):
        if callable(self._on_state_change):
            await self._on_state_change(self.state)

    async def _do_on_message(self, message: str):
        if callable(self._on_message):
            await self._on_message(message)

    async def _do_on_connect(self):
        if callable(self._on_connect):
            await self._on_connect()

    async def _on_message_runner(self):
        while True:
            message = await self._incoming_queue.get()
            await self._do_on_message(message)
            self._incoming_queue.task_done()

    async def send(self, message: str):
        """Add message to send queue"""
        if not isinstance(message, str):
            raise ValueError(f"Unsupported message type {type(message)}")
        await self._outgoing_queue.put(message)

    async def runner(self):
        """Send and receive messages on websocket"""

        async def consumer(ws: ClientConnection):
            """Enqueue messages received on websocket"""
            try:
                async for message in ws:
                    if isinstance(message, str):
                        _LOGGER.debug("Message received %s", message)
                        await self._incoming_queue.put(message)
            except asyncio.CancelledError:
                _LOGGER.debug("Consumer was cancelled")
                raise

        async def producer(ws: ClientConnection):
            """Send queued messages on websocket"""
            try:
                while True:
                    message = await self._outgoing_queue.get()
                    _LOGGER.debug("Sending message %s", message)
                    await ws.send(message)
                    self._outgoing_queue.task_done()
            except asyncio.CancelledError:
                _LOGGER.debug("Producer cancelled")
                raise

        async def websocket_handler(websocket: ClientConnection):
            """Websocket handler"""
            with task_manager(cancel_on_exit=True) as tasks:
                try:
                    tasks.append(
                        asyncio.create_task(consumer(websocket), name="consumer")
                    )
                    tasks.append(
                        asyncio.create_task(producer(websocket), name="producer")
                    )
                finally:
                    await tasks.wait(return_when=asyncio.FIRST_COMPLETED)

        uri = f"ws://{self._host}:{self._port}"
        while True:
            try:
                _LOGGER.info("Connecting to %s", uri)
                async for websocket in connect(uri, open_timeout=self._connect_timeout):
                    try:
                        self._ws = websocket
                        _LOGGER.info("Connection established to %s", uri)
                        await self._do_on_connect()
                        await self._do_on_state_change()
                        await websocket_handler(
                            websocket
                        )  # this will return if connection is closed OK by remote
                        _LOGGER.warning(
                            "Connection to %s was closed, will reconnect", uri
                        )
                        await self._do_on_state_change()
                    except Exception as e:  # pylint: disable=W0718
                        try:
                            await self._do_on_state_change()
                            processed_exception = process_exception(e)
                            if not process_exception:
                                # transient error
                                _LOGGER.error(
                                    "Failed to connect to %s, will retry",
                                    uri,
                                    exc_info=1,
                                )
                                continue
                            else:
                                raise processed_exception  # pylint: disable=W0707
                        except ConnectionClosed:
                            _LOGGER.warning(
                                "Connection to %s was closed unexpectedly, will retry",
                                uri,
                            )
                            # reconnect if connection closed
                            continue

            except asyncio.CancelledError:
                _LOGGER.info("Shutting down connection to %s", uri)
                await self._do_on_state_change()
                raise

    def connect(self):
        """Connect to server"""
        if self._ws is None:
            runner_task = asyncio.create_task(self.runner(), name="runner")
            message_runner_task = asyncio.create_task(self._on_message_runner())
            self._tasks.append(runner_task, message_runner_task)

    def close(self):
        """Close connection and perform clean up"""
        self._tasks.cancel()
        self._ws = None

    def __enter__(self):
        """Initiate connection"""
        self.connect()

    def __exit__(self, *args, **kwargs):
        self.close()



"""
    Configure testing
"""


import pytest_asyncio

from .utils.test_server import TestServer


@pytest_asyncio.fixture()
async def ws_server():
    """Websocket test server"""
    async with TestServer("localhost", "3001") as server:
        yield server
        await server.close()



from __future__ import annotations


class ErrorCache:
    """Error cache, caches previous data until frame is complete"""

    def __init__(self) -> None:
        self._current: list[str] = []
        self._next: list[str] = []
        self._is_collecting = False

    @property
    def data(self):
        """Return current error messages

        :return: error messages
        :rtype: list[str]
        """
        return self._current

    def add(self, message: str):
        """Add error message

        :param message: error message
        :type message: str
        """
        if self._is_collecting:
            self._next.append(message)

    def end_frame(self):
        """Mark data frame as complete"""
        self._current = self._next
        self._next = []
        self._is_collecting = False

    def begin_frame(self):
        """Begin new frame"""
        self._is_collecting = True



"""HRV System client"""

import asyncio
import logging
from typing import Callable

from websockets.protocol import State

from .const import DataKey, MessageType
from .data import IncomingMessage, OutgoingMessage, ParseError
from .helpers.error_cache import ErrorCache
from .helpers.websocket import ReconnectingWebsocketClient

_LOGGER: logging.Logger = logging.getLogger(__name__)


class Client:
    """Client to manage communication with HRV"""

    def __init__(
        self, ip: str, port: int = 3001, update_interval=30, connect_timeout=15
    ):
        """Initiate client

        :param ip: ip address of the unit
        :type ip: str
        :param port: port
        :type port: int
        :param update_interval: update interval for calling handler, defaults to 30
        :type update_interval: int, optional
        :param connect_timeout: timeout when establishing connection, defaults to 15
        :type connect_timeout: int, optional
        """
        self._update_interval = update_interval
        self._ip = ip
        self._port = port
        self._data: dict[DataKey, str] = {}
        self._error_cache = ErrorCache()
        self._on_message_handlers: set[Callable[[dict[DataKey, str]]]] = set()
        self._on_state_change_handlers: set[Callable[[dict[DataKey, str]]]] = set()
        self._connect_timeout = connect_timeout
        self._tasks = [asyncio.create_task(self._do_call_message_handlers())]
        self._websocket = ReconnectingWebsocketClient(
            host=self._ip,
            port=self._port,
            connect_timeout=self._connect_timeout,
            on_message=self._on_message,
            on_connect=self._send_start_message,
            on_state_change=self._on_state_change,
        )

    @property
    def state(self):
        """State of the underlying websocket connection"""
        return self._websocket.state

    @property
    def data(self):
        """Get data from system if connection is alive"""
        if self.state == State.OPEN:
            return self._data

        return dict()

    def connect(self):
        """Connect to HRV and begin receiving"""
        self._websocket.connect()

    async def _send_start_message(self):
        """Send start message to server to begin receiving data"""
        await self._websocket.send("#:\r")

    async def _do_call_message_handlers(self):
        """Call message handlers with data at update_interval"""
        while True:
            await asyncio.sleep(self._update_interval)
            self._call_message_handlers()

    def _call_message_handlers(self):
        """Call handlers with data"""
        for handler in self._on_message_handlers:
            try:
                handler(self.data)
            except Exception:
                _LOGGER.error("Failed to call handler %s", handler, exc_info=1)

    def _call_state_change_handlers(self, state):
        """Call handlers with data"""
        for handler in self._on_state_change_handlers:
            try:
                handler(state)
            except Exception:
                _LOGGER.error("Failed to call handler %s", handler, exc_info=1)

    def close(self):
        """Disconnect from system"""
        try:
            pass
        finally:
            if self._websocket:
                self._websocket.close()
            for task in self._tasks:
                task.cancel()

    async def _on_state_change(self, state):
        self._call_state_change_handlers(state)

    async def _on_message(self, msg: str):
        """Update data"""
        try:
            (key, value, message_type) = IncomingMessage.from_str(msg)

            if key in [
                DataKey.ERROR_FRAME_START,
                DataKey.ERROR_MESSAGE,
                DataKey.ERROR_FRAME_END,
            ]:
                if key == DataKey.ERROR_FRAME_START:
                    self._error_cache.begin_frame()
                if key == DataKey.ERROR_MESSAGE:
                    self._error_cache.add(value)
                if key == DataKey.ERROR_FRAME_END:
                    self._error_cache.end_frame()
                    self._data[DataKey.ERROR_MESSAGE.value] = self._error_cache.data
            else:
                self._data[key] = value
                if message_type == MessageType.ACK_OK:
                    self._call_message_handlers()
        except ParseError as e:
            _LOGGER.warning(e, exc_info=1)

    def add_state_change_handler(self, handler: Callable[[str], None]):
        """Add state change handler to be called when client state changes

        :param handler: handler to be added
        :type handler: Callable[[str], None]
        """
        self._on_state_change_handlers.add(handler)

    def remove_state_change_handler(self, handler: Callable[[str], None]):
        """Remove state change handler

        :param handler: handler to be removed
        :type handler: Callable[[str], None]
        """
        self._on_state_change_handlers.remove(handler)

    def add_message_handler(self, handler: Callable[[str], None]):
        """Add message handler

        Message handler will be called at update interval

        :param handler: handler function. Must be safe to call from event loop
        :type handler: Callable[[str], None]
        """
        self._on_message_handlers.add(handler)

    def remove_message_handler(self, handler: Callable[[str], None]):
        """Remove message handler

        :param handler: handler to remove
        :type handler: Callable[[str], None]
        """
        self._on_message_handlers.remove(handler)

    async def send_command(self, key: MessageType, payload: str | int):
        """Send command to HRV unit

        :param key: message type key
        :type key: MessageType
        :param payload: payload
        :type value: str | int
        """
        message = OutgoingMessage(key, payload)

        async def ack_command():
            """Should probably ack command here, just sleep for now"""
            await asyncio.sleep(0.5)

        await self._websocket.send(str(message))
        await asyncio.gather(ack_command())

    def __enter__(self, *args, **kwargs):
        self.connect()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()



"""Client tests"""
import asyncio
import logging
import typing

import pytest
import pytest_asyncio
from websockets.protocol import State

from pysaleryd.client import Client

if typing.TYPE_CHECKING:
    from utils.test_server import TestServer

__author__ = "Björn Dalfors"
__copyright__ = "Björn Dalfors"
__license__ = "MIT"


async def has_state(client: Client, state: State):
    try:
        while client.state != state:
            await asyncio.sleep(1)
    except asyncio.CancelledError:
        pass
    finally:
        assert client.state == state


@pytest_asyncio.fixture(name="hrv_client")
async def _hrv_client(ws_server: "TestServer"):
    """HRV Client"""
    with Client("localhost", 3001, 3, connect_timeout=3) as client:
        yield client


@pytest.mark.asyncio
async def test_client_connect(hrv_client: "Client"):
    """test connect"""
    await has_state(hrv_client, State.OPEN)


@pytest.mark.asyncio
async def test_handler(hrv_client: "Client", mocker, caplog):
    caplog.set_level(logging.DEBUG)
    """Test handler callback"""

    data = None  # noqa: F841

    def handler(_data):
        nonlocal data
        data = _data

    def broken_handler(data):
        raise Exception()  # pylint: disable=W0719

    hrv_client.add_message_handler(broken_handler)
    hrv_client.add_message_handler(handler)
    await asyncio.sleep(5)

    assert isinstance(data, dict)
    assert any(data.keys())


@pytest.mark.asyncio
async def test_state_change_handler(
    hrv_client: "Client", ws_server: "TestServer", mocker, caplog
):
    caplog.set_level(logging.DEBUG)
    """Test state change handler callback"""

    class Foo:
        def handler(self, state: State):
            pass

    foo = Foo()
    spy = mocker.spy(foo, "handler")
    hrv_client.add_state_change_handler(foo.handler)
    await asyncio.sleep(1)
    spy.assert_called_once_with(State.OPEN)
    await ws_server.close()
    await asyncio.sleep(1)
    spy.assert_called_with(State.CLOSED)
    await ws_server.start()
    await asyncio.sleep(5)
    spy.assert_called_with(State.OPEN)


@pytest.mark.asyncio
async def test_get_data(hrv_client: "Client", caplog):
    """Test get data"""
    caplog.set_level(logging.DEBUG)
    await asyncio.sleep(5)
    assert isinstance(hrv_client.data, dict)
    assert any(hrv_client.data.keys())


@pytest.mark.asyncio
async def test_reconnect(hrv_client: "Client", ws_server: "TestServer", caplog):
    """Test reconnect"""
    caplog.set_level(logging.DEBUG)

    await asyncio.wait_for(has_state(hrv_client, State.OPEN), 15)
    await ws_server.close()
    await asyncio.wait_for(has_state(hrv_client, State.CLOSED), 15)
    await ws_server.start()
    await asyncio.wait_for(has_state(hrv_client, State.OPEN), 15)


@pytest.mark.asyncio
async def test_connect_unresponsive(ws_server: "TestServer", caplog):
    """Test connection is retried when host is unresponsive"""
    caplog.set_level(logging.INFO)

    await ws_server.close()
    await asyncio.sleep(1)
    client = Client("localhost", 3001, 3, 1)
    client.connect()
    await asyncio.sleep(1)
    await ws_server.start()
    await asyncio.wait_for(has_state(client, State.OPEN), 15)


@pytest.mark.asyncio
async def test_send_command(hrv_client: "Client"):
    """Test send command"""
    await hrv_client.send_command("MF", 0)


@pytest.mark.asyncio
async def test_disconnect(hrv_client: "Client", caplog):
    caplog.set_level(logging.DEBUG)
    """Test disconnected client remains disconnected"""
    await has_state(hrv_client, State.OPEN)
    await asyncio.sleep(1)
    hrv_client.close()
    await asyncio.sleep(5)
    await has_state(hrv_client, None)



import logging

import pytest

from pysaleryd.const import DataKey, MessageType
from pysaleryd.data import IncomingMessage, ParseError, SystemProperty

__author__ = "Björn Dalfors"
__copyright__ = "Björn Dalfors"
__license__ = "MIT"


_LOGGER = logging.getLogger(__name__)


def test_parse_int_from_list_str():
    """Test parsing int list"""
    (key, value, message_type) = IncomingMessage.from_str("#MF: 1+ 0+ 2+30\r")
    assert key == DataKey.MODE_FAN
    assert isinstance(value, str)
    assert message_type == MessageType.MESSAGE


def test_parse_ack():
    (key, value, message_type) = IncomingMessage.from_str("#$MF:  1+  0+  2+0\r")
    assert key == "MF"
    assert isinstance(value, str)
    assert message_type == MessageType.ACK_OK


def test_parse_ack_error():
    (key, value, message_type) = IncomingMessage.from_str("#!MF:\r")
    assert key == "MF"
    assert isinstance(value, str)
    assert message_type == MessageType.ACK_ERROR


def test_parse_int_from_str():
    """Test parsing int"""
    (key, value, message_type) = IncomingMessage.from_str("#*XX:0\r")
    assert key == "*XX"
    assert isinstance(value, str)
    assert value == "0"
    assert message_type == MessageType.MESSAGE


def test_parse_str_from_str():
    """Test parsing str"""
    (key, value, message_type) = IncomingMessage.from_str("#*XX:xxx\r")
    assert key == "*XX"
    assert isinstance(value, str)
    assert value == "xxx"
    assert message_type == MessageType.MESSAGE

    (key, value, message_type) = IncomingMessage.from_str("#*XX:    1.x.1\r")
    assert key == "*XX"
    assert isinstance(value, str)
    assert value == "1.x.1"
    assert message_type == MessageType.MESSAGE


def test_parse_error():
    """Test parse error"""
    with pytest.raises(ParseError):
        IncomingMessage.from_str("wer")


def test_parse_system_property():
    """Test parse SystemProperty"""
    key = DataKey.INSTALLER_PASSWORD
    value_str = "1"
    parsed = SystemProperty.from_str(key, value_str)
    assert parsed.key == DataKey.INSTALLER_PASSWORD
    assert parsed.value == 1

    value_str = "1+ 2+ 3+0"
    parsed = SystemProperty.from_str(key, value_str)
    assert parsed.value == 1
    assert parsed.min_value == 2
    assert parsed.max_value == 3

    value_str = "test"
    parsed = SystemProperty.from_str(key, value_str)
    assert parsed.value == "test"



#!/usr/bin/env python

import asyncio
import logging

from websockets.asyncio.server import ServerConnection, serve
from websockets.exceptions import ConnectionClosed

from pysaleryd.helpers.task import task_manager

_LOGGER = logging.getLogger(__name__)


async def data_generator(ws: ServerConnection):
    """Generate data and push to queue"""
    while True:
        try:
            await ws.send("#MF: 1+ 1+ 1+1")
            await asyncio.sleep(0.5)
        except ConnectionClosed:
            break


class TestServer:
    """Test server"""

    async def _handler(self, websocket: ServerConnection):
        """Websocket handler to emulate HRV"""
        with task_manager(cancel_on_exit=True) as task_list:
            message = await websocket.recv()
            _LOGGER.debug("Received %s", message)
            task = self._loop.create_task(
                data_generator(websocket), name="data_generator"
            )
            task_list.append(task)
            await task_list.wait(return_when=asyncio.ALL_COMPLETED)

    def __init__(self, host, port, loop=None):
        self.port = port
        self.host = host
        self._stop = None
        self._server: asyncio.Server = None
        self._gen = None
        self._loop = loop if loop is not None else asyncio.get_running_loop()

    async def close(self):
        """Stop server"""
        if self._server:
            self._server.close()
            await self._server.wait_closed()

    async def start(self):
        """Start server"""
        self._server = await serve(
            self._handler, self.host, self.port, start_serving=True
        )

    async def __aenter__(self, *args, **kwargs):
        await self.start()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.close()



"""Utils"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from .const import MessageType, PayloadSeparator

if TYPE_CHECKING:
    from typing import Optional, Union

    from .const import DataKey

_LOGGER: logging.Logger = logging.getLogger(__package__)


class ParseError(BaseException):
    """Parse error. Raised when message parsing fails"""


@dataclass
class OutgoingMessage:
    """Outgoing message to system"""

    key: DataKey
    value: str

    def __str__(self):
        return f"#{self.key}:{self.value}\r"


class IncomingMessage:
    """Incoming message from system"""

    @staticmethod
    def from_str(msg: str) -> tuple[DataKey, str, MessageType]:
        """Parse message string

        Args:
            msg (str): raw message

        Raises:
            ParseError: if parsing fails

        Returns:
            (key, value, message_type): parsed message key and value
        """
        if msg[0] != PayloadSeparator.MESSAGE_START:
            raise ParseError(f"Unsupported payload, {msg}")
        else:
            msg = msg[1::]
        try:
            [key, payload] = [
                v.strip() for v in msg.split(PayloadSeparator.PAYLOAD_START, 1)
            ]

            if key[0] in [PayloadSeparator.ACK_ERROR, PayloadSeparator.ACK_OK]:
                return (
                    key[1::],
                    payload,
                    MessageType.ACK_OK.value
                    if key[0] == PayloadSeparator.ACK_OK
                    else MessageType.ACK_ERROR.value,
                )
            else:
                return (key, payload, MessageType.MESSAGE.value)
        except Exception as exc:
            raise ParseError(f"Failed to parse message {msg}") from exc


class BaseSystemProperty:
    """HRV System property"""


class SystemProperty(BaseSystemProperty):
    """HRV System property with value, min, max"""

    def __init__(
        self,
        key: DataKey,
        value: Optional[int | str] = None,
        min_value: Optional[int | str] = None,
        max_value: Optional[int | str] = None,
        *_args,
    ):
        self.key = key
        self.value = value
        self.min_value = min_value
        self.max_value = max_value

    @classmethod
    def from_str(cls, key: DataKey, raw_value: str):
        """Create instance from from string"""

        def maybe_cast(x: str) -> Union[int, float, str, None]:
            """Cast value if it is numeric"""
            if x is None:
                return x
            if x.isdigit():
                return int(x)
            if x.replace(".", "", 1).isdigit():
                return float(x)
            return x

        [*positions] = (
            [maybe_cast(v.strip()) for v in raw_value.split("+")]
            if raw_value is not None
            else []
        )

        return cls(key, *positions)


class ErrorSystemProperty(BaseSystemProperty):
    """HRV System error property"""

    def __init__(self, key: DataKey, value: Optional[list[str]] = None):
        self.key = key
        self.value = value



"""Constants"""

from enum import IntEnum, StrEnum


class MessageType(IntEnum):
    """Message type"""

    ACK_OK = 0
    ACK_ERROR = 1
    MESSAGE = 2
    RAW = 3


class PayloadSeparator(StrEnum):
    """Message payload separator"""

    ACK_OK = "$"
    ACK_ERROR = "!"
    MESSAGE_START = "#"
    PAYLOAD_START = ":"


class DataKey(StrEnum):
    """Data key enumerator"""

    # Scalar
    AIR_TEMPERATURE_AT_HEATER = "*TK"
    AIR_TEMPERATURE_SUPPLY = "*TC"
    CONTROL_SYSTEM_VERSION = "*SC"
    ERROR_FRAME_END = "*EZ"
    ERROR_MESSAGE = "*EB"
    ERROR_FRAME_START = "*EA"
    FAN_SPEED_EXHAUST = "*DB"
    FAN_SPEED_SUPPLY = "*DA"
    FILTER_MONTHS_LEFT = "*FL"
    HEAT_EXCHANGER_ROTOR_PERCENT = "*XA"
    HEAT_EXCHANGER_ROTOR_RPM = "*XB"
    HEATER_POWER_PERCENT = "*MJ"
    MINUTES_LEFT_BOOST_MODE = "*FI"
    MINUTES_LEFT_FIREPLACE_MODE = "*ME"
    MODEL_NAME = "*SB"
    PRODUCT_NUMBER = "*SA"
    INSTALLER_EMAIL = "IB"
    INSTALLER_NAME = "IA"
    INSTALLER_PASSWORD = "IP"
    INSTALLER_PHONE = "IC"
    INSTALLER_WEBSITE = "IE"

    # Vecto x+ x+ x+x
    CONTROL_SYSTEM_STATE = "MP"
    MODE_HEATER_POWER_RATING = "MG"
    COOLING_MODE = "MK"
    FIREPLACE_MODE = "MB"
    MODE_FAN = "MF"
    MODE_TEMPERATURE = "MT"
    TARGET_TEMPERATURE_COOL = "TF"
    TARGET_TEMPERATURE_ECONOMY = "TE"
    TARGET_TEMPERATURE_NORMAL = "TD"
    MODE_HEATER = "MH"
